#!/usr/bin/env python
# @(#) $Id: check_acld.py 654 2010-03-05 23:34:23Z leres $ (LBL)
#
# Copyright (c) 2009, 2010
#	The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#

import optparse
import os
import re
import subprocess
import sys

# Globals
prog = os.path.basename(sys.argv[0])

STR2STATUS = {
    'OK':	0,
    'WARNING':	1,
    'CRITICAL':	2,
    'UNKNOWN':	3,
}

STATUS2STR = {
    0: 'OK',
    1: 'WARNING',
    2: 'CRITICAL',
    3: 'UNKNOWN',
}

# Adjust our path
pathstr = 'PATH'
acldbin = '/usr/operator/acl2/scripts'
if not acldbin in os.environ[pathstr].split(':'):
    os.environ[pathstr] = '%s:%s' % (acldbin, os.environ[pathstr])

def checkdroprestore(data):
    """Test drop/restore"""
    prefix = ', Test Drop/Restore: '
    for addr in data['args']:
	result = ''
	for what in ['drop', 'restore']:
	    data['sout'] = ''
	    data['serr'] = ''
	    runacl(data, [what, addr])
	    if data['code'] != 0:
		data['status'] = STR2STATUS['CRITICAL']
		result = '!'
		break
	data['msg'] += prefix
	prefix = ', '
	data['msg'] += addr + result

def checkpct(data):
    """Check if any ACLs have high sequence numbers"""
    if 'pct' not in data:
	return

    m = re.search(r'^seqrange (\d+) (\d+)\n', data['sout'], re.MULTILINE)
    if m == None:
	data['status'] = STR2STATUS['UNKNOWN']
	data['msg'] += " Can't find seqrange"
	return
    seqrange = (int(m.group(1)), int(m.group(2)))
    size = seqrange[1] - seqrange[0]
    if size < 0:
	data['status'] = STR2STATUS['UNKNOWN']
	data['msg'] += " Can't figure out seqrange size"
	return

    re_lastseq = re.compile(r'\A# lastseq for ACL (.+) is (\d+)\Z')

    pct = data['pct']
    lines = data['sout'].split('\n')
    acls = []
    for line in lines:
	m = re_lastseq.search(line)
	if m == None:
	    continue
	nm = m.group(1)
	lastseq = int(m.group(2))
	n = lastseq - seqrange[0]
	if n < 0:
	    n = 0
	v = (100 * n) / size
	acl = [nm, v]
	acls.append(acl)

    # XXX leave in config order?
    #acls.sort()

    prefix = '; Sequence number PCTs: '
    for acl in acls:
	data['msg'] += prefix
	prefix = ', '
	data['msg'] += '%s=%d%%' % tuple(acl)
	if pct > 0 and acl[1] >= pct:
	    data['status'] = STR2STATUS['CRITICAL']
	    data['msg'] += '!'

def checkrate(data):
    """Check if the rate is ok"""
    if 'rate' not in data:
	return
    rate = data['rate']
    lines = data['sout'].split('\n')
    re_rate = re.compile(r'\A# rate for ACL (.+) is (\d+) acl/min\Z')
    acls = []
    for line in lines:
	m = re_rate.search(line)
	if m == None:
	    continue
	acl = [m.group(1), int(m.group(2))]
	acls.append(acl)

    # XXX leave in config order?
    #acls.sort()

    prefix = ', Drop rates: '
    for acl in acls:
	data['msg'] += prefix
	prefix = ', '
	data['msg'] += '%s=%d' % tuple(acl)
	if rate > 0 and acl[1] >= rate:
	    data['status'] = STR2STATUS['CRITICAL']
	    data['msg'] += '!'
    data['msg'] += ' acl/min'

def checkstate(data):
    """Check if the state is ok, return a dict containing information"""

    # Extract host
    m = re.search(r'^# host (.+)\n', data['sout'], re.MULTILINE)
    if m == None:
	data['status'] = STR2STATUS['UNKNOWN']
	data['msg'] = "Can't determine acld host"
	return
    data['host'] = m.group(1)

    # Extract state
    m = re.search(r'^# state (.+)\n', data['sout'], re.MULTILINE)
    if m == None:
	data['status'] = STR2STATUS['UNKNOWN']
	data['msg'] = "can't find state"
	return
    data['state'] = m.group(1)

    if data['state'] != 'loggedin':
	data['status'] = STR2STATUS['CRITICAL']
	data['msg'] = 'acld on %s not logged in' % (data['host'])
    data['msg'] = 'acld on %s logged in and ready' % (data['host'])

def reportandexit(data):
    """Produce the nagios report and then exit with the appropriate code"""
    status = data['status']
    print '%s: %s' % (STATUS2STR[status], data['msg'])
    sys.exit(status)

def runacl(data, args):
    """Check the status of an acl result"""
    cmd = ['acl.exp' ]

    # Set up environment
    if data['addr'] != None:
	estr = 'ACLIPADDR'
	if estr in os.environ:
	    oldaddrenv = os.environ[estr]
	else:
	    oldaddrenv = None
	os.environ[estr] = data['addr']

    if data['port'] != None:
	estr = 'ACLPORT'
	if estr in os.environ:
	    oldportenv = os.environ[estr]
	else:
	    oldportenv = None
	os.environ[estr] = unicode(data['port'])

    cmd.extend(args)
    child = subprocess.Popen(cmd,
	stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    (sout, serr) = child.communicate()
    # Restore environment
    if data['addr'] != None:
	estr = 'ACLIPADDR'
	if oldaddrenv != None:
	    os.environ[estr] = oldaddrenv
	else:
	    del os.environ[estr]
    data['code'] = child.returncode

    if data['port'] != None:
	estr = 'ACLPORT'
	if oldportenv != None:
	    os.environ[estr] = oldportenv
	else:
	    del os.environ[estr]
    data['serr'] = serr
    data['sout'] = sout

def main():
    """Main program"""

    usage = 'usage: %s [-a addr] [-H pct] [-p port] [-R rate] addr ...' % (prog)

    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-a', None,
	    action='store', dest='addr',
		    help='IP address of acld')
    parser.add_option('-p', None,
	    action='store', dest='port', type='int',
		    help='TCP port of acld')
    parser.add_option('-R', None,
	    action='store', dest='rate', default=0,
		    help='ACL drop rate to report critical')
    parser.add_option('-H', None,
	    action='store', dest='pct', type='int',
		    help='lastseq highwater percentage to report critical')

    (opts, args) = parser.parse_args()

    data = {
	'addr': opts.addr,
	'args': args,
	'code': -1,
	'msg': '',
	'pct': opts.pct,
	'port': opts.port,
	'rate': int(opts.rate),
	'serr': '',
	'sout': '',
	'status': STR2STATUS['OK'],
    }

    runacl(data, ['state'])

    if data['code'] != 0:
	data['status'] = STR2STATUS['UNKNOWN']
	data['msg'] = 'acl.exp state failed'
	reportandexit(data)

    # Are we logged in?
    checkstate(data)

    # Are any drop rates too high?
    checkrate(data)

    # Are any sequence numbers getting to high?
    checkpct(data)

    # Are any drop rates too high (run this test last)
    checkdroprestore(data)

    # That's all, folks!
    reportandexit(data)

if __name__ == "__main__":
    main()
