/*
 * Copyright (c) 2002, 2003, 2004, 2006, 2007, 2008, 2009, 2010
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: client.c 650 2010-02-04 03:02:03Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "acld.h"
#include "cf.h"
#include "child.h"

/* Globals */
struct s2v cmd2req[] = {
	{ "reload",		REQ_RELOAD },
	{ "drop",		REQ_DROP },
	{ "blockhosthost",	REQ_BLOCKHOSTHOST },
	{ "restore",		REQ_RESTORE },
	{ "restorehosthost",	REQ_RESTOREHOSTHOST },
	{ "dropudpport",	REQ_DROPUDPPORT },
	{ "restoreudpport",	REQ_RESTOREUDPPORT },
	{ "droptcpport",	REQ_DROPTCPPORT },
	{ "restoretcpport",	REQ_RESTORETCPPORT },
	{ "permitudpdsthostport", REQ_PERMITUDPDSTHOSTPORT },
	{ "unpermitudpdsthostport", REQ_UNPERMITUDPDSTHOSTPORT },
	{ "permittcpdsthostport", REQ_PERMITTCPDSTHOSTPORT },
	{ "unpermittcpdsthostport", REQ_UNPERMITTCPDSTHOSTPORT },
	{ "droptcpdsthostport", REQ_DROPTCPDSTHOSTPORT },
	{ "restoretcpdsthostport", REQ_RESTORETCPDSTHOSTPORT },
	{ "nullzero",		REQ_NULLZERO },
	{ "nonullzero",		REQ_NONULLZERO },
	{ "state",		REQ_STATE },
	{ "listacl",		REQ_LISTACL },
	{ "listroute",		REQ_LISTROUTE },
	{ "whitelist",		REQ_WHITELIST },
	{ "addwhitelist",	REQ_ADDWHITELIST },
	{ "remwhitelist",	REQ_REMWHITELIST },
	{ "query",		REQ_QUERY },
	{ "querynullzero",	REQ_QUERYNULLZERO },
	{ "querywhitelist",	REQ_QUERYWHITELIST },
	{ "compact",		REQ_COMPACT },
	{ "exit",		REQ_EXIT },
	{ "test",		REQ_TEST },		/* undocumented */
	{ "help",		REQ_HELP },
	{ "unknown",		REQ_UNKNOWN },
	{ NULL,			REQ_UNKNOWN }
};

struct s2v ctype2str[] = {
	{ "priv",		CTYPE_PRIV },
	{ "ro",			CTYPE_RO },
	{ "web",		CTYPE_WEB },
	{ NULL,			0 }
};

/* Forwards */
int clientallowed(struct client *, struct req *);
void clientcfstate(struct cf *, struct req *);
int clientcomment(struct req *, struct iobuf *);
void clientcompact(struct state *, struct client *, struct req *);
void clienthelp(struct client *, struct req *);
void clientlistacl(struct state *, struct client *, struct req *);
void clientlistroute(struct state *, struct client *, struct req *);
void clientquery(struct state *, struct client *, struct req *);
void clientquerynullzero(struct state *, struct client *, struct req *);
void clientquerywhitelist(struct state *, struct client *, struct req *);
void clientstate(struct state *, struct client *, struct req *);
void clientwhitelist(struct state *, struct client *, struct req *);

static enum reqtype ro_ok[] = {
	REQ_UNKNOWN,
	REQ_STATE,
	REQ_LISTACL,
	REQ_LISTROUTE,
	REQ_WHITELIST,
	REQ_QUERY,
	REQ_QUERYNULLZERO,
	REQ_QUERYWHITELIST,
	REQ_TEST,
	REQ_EXIT,
	REQ_HELP
};

/* Check if a client is allowed to execute this command */
int clientallowed(struct client *cl, struct req *rp)
{
	int i, n;
	const char *port;

	if (cl->type == CTYPE_PRIV)
		return (1);

	/* Handle readonly guys */
	for (i = 0, n = sizeof(ro_ok) / sizeof(ro_ok[0]); i < n; ++i)
		if (ro_ok[i] == rp->type)
			return (1);

	/* The only other client type is web registration */
	if (cl->type != CTYPE_WEB)
		return (0);

	/* These are the only additional allowed requests */
	if (rp->type != REQ_PERMITTCPDSTHOSTPORT &&
	    rp->type != REQ_UNPERMITTCPDSTHOSTPORT)
		return (0);

	/* Look for right number of arguments */
	if (rp->an != 4)
		return (0);

	/* Only two ports are allowed */
	port = rp->av[3];
	if (strcmp(port, "80") == 0 || strcmp(port, "443") == 0)
		return (1);

	/* Deny anything we don't understand */
	return (0);
}

void
clientfinishreq(struct state *sp, struct client *cl, struct req *rp)
{
	long v;
	char *cp, *ep;
	const char *p;

	/* Only say something back if there was a command */
	if (rp->an < 1) {
		freereq(rp);
		return;
	}

	/* Parse command */
	rp->type = str2val(cmd2req, rp->av[0]);
	rp->cmd = val2str(cmd2req, rp->type);
	if (rp->type == REQ_UNKNOWN) {
		clientsenderr(cl, rp, "unknown command \"%s\"", rp->av[0]);
		freereq(rp);
		return;
	}

	/* We need a minimum of a command and a cookie */
	if (rp->an < 2) {
		clientsenderr(cl, rp, "missing cookie");
		freereq(rp);
		return;
	}

	/* Parse cookie */
	rp->cookie = strtol(rp->av[1], &ep, 10);
	if (*ep != '\0') {
		clientsenderr(cl, rp, "non-numeric cookie");
		freereq(rp);
		return;
	}

	/* Enforce client privileges */
	if (!clientallowed(cl, rp)) {
		lg(LOG_ERR,
		    "clientinput: client #%d: %s request denied for %s client",
		    cl->n, val2str(cmd2req, rp->type),
		    val2str(ctype2str, cl->type));
		clientsenderr(cl, rp, "Permission denied");
		freereq(rp);
/* XXX need to add this? cl->treq = NULL  */
		return;
	}

	/* Verify parameters and build request */
	switch (rp->type) {

	case REQ_EXIT:
	case REQ_HELP:
	case REQ_LISTROUTE:
	case REQ_RELOAD:
	case REQ_STATE:
	case REQ_TEST:
	case REQ_WHITELIST:
		if (rp->an != 2) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}
		break;

	case REQ_DROP:
	case REQ_RESTORE:
	case REQ_QUERY:
		if (rp->an < 3 || rp->an > 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		/* Extract host or network addr */
		if (rp->an == 4)
			cp = rp->av[3];
		else
			cp = NULL;
		p = extractaddr(rp->av[2], cp, &rp->acl.addr1);
		if (p != NULL) {
			clientsenderr(cl, rp, "extractaddr: %s", p);
			freereq(rp);
			return;
		}

		p = checkmaskwidth(&rp->acl.addr1, MAXV4_SIZE, MAXV6_SIZE);
		if (p != NULL) {
			clientsenderr(cl, rp, "%s", p);
			freereq(rp);
			return;
		}

		if (ISHOST(&rp->acl.addr1))
			rp->acl.type = ATYPE_BLOCKHOST;
		else
			rp->acl.type = ATYPE_BLOCKNET;
		break;

	case REQ_BLOCKHOSTHOST:
	case REQ_RESTOREHOSTHOST:
		if (rp->an != 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		/* Extract host addrs */
		if ((p = extracthost(rp->av[2], NULL,
		    &rp->acl.addr1)) != NULL ||
		    (p = extracthost(rp->av[3], NULL,
		    &rp->acl.addr2)) != NULL) {
			clientsenderr(cl, rp, "extracthost: %s", p);
			freereq(rp);
			return;
		}

		rp->acl.type = ATYPE_BLOCKHOSTHOST;
		break;

	case REQ_DROPUDPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
		if (rp->an != 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		if (sp->cf->c_lowportseq == 0 || sp->cf->c_highportseq == 0) {
			clientsenderr(cl, rp,
			    "portseqrange are not configured");
			freereq(rp);
			return;
		}

		rp->acl.port1 = strsave(rp->av[2]);
		rp->aclname = strsave(rp->av[3]);
		if (rp->type == REQ_DROPUDPPORT ||
		    rp->type == REQ_RESTOREUDPPORT)
			rp->acl.type = ATYPE_BLOCKUDPPORT;
		else
			rp->acl.type = ATYPE_BLOCKTCPPORT;
		break;

	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
		if (rp->an != 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		if (sp->cf->c_lowpermithostportseq == 0 ||
		    sp->cf->c_highpermithostportseq == 0) {
			clientsenderr(cl, rp,
			    "permithostportseqrange are not configured");
			freereq(rp);
			return;
		}

		/* Extract host addr */
		p = extracthost(rp->av[2], NULL, &rp->acl.addr1);
		if (p != NULL) {
			clientsenderr(cl, rp, "extractaddr: %s", p);
			freereq(rp);
			return;
		}

		rp->acl.port1 = strsave(rp->av[3]);
		rp->acl.type = ATYPE_PERMITUDPDSTHOSTPORT;
		break;

	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
		if (rp->an != 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		if (sp->cf->c_lowpermithostportseq == 0 ||
		    sp->cf->c_highpermithostportseq == 0) {
			clientsenderr(cl, rp,
			    "permithostportseqrange are not configured");
			freereq(rp);
			return;
		}

		/* Extract host addr */
		p = extracthost(rp->av[2], NULL, &rp->acl.addr1);
		if (p != NULL) {
			clientsenderr(cl, rp, "extractaddr: %s", p);
			freereq(rp);
			return;
		}

		rp->acl.port1 = strsave(rp->av[3]);
		rp->acl.type = ATYPE_PERMITTCPDSTHOSTPORT;
		break;

	case REQ_DROPTCPDSTHOSTPORT:
	case REQ_RESTORETCPDSTHOSTPORT:
		if (rp->an != 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		/* Extract host addr */
		p = extracthost(rp->av[2], NULL, &rp->acl.addr1);
		if (p != NULL) {
			clientsenderr(cl, rp, "extractaddr: %s", p);
			freereq(rp);
			return;
		}

		rp->acl.port1 = strsave(rp->av[3]);
		rp->acl.type = ATYPE_BLOCKTCPDSTHOSTPORT;
		break;

	case REQ_NULLZERO:
	case REQ_NONULLZERO:
	case REQ_QUERYNULLZERO:
		if (rp->an != 3) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		/* Extract host addr */
		p = extracthost(rp->av[2], NULL, &rp->nullzero.dst);
		if (p != NULL) {
			clientsenderr(cl, rp, "extractaddr: %s", p);
			freereq(rp);
			return;
		}

		rp->nullzero.type = ROUTE_NULLZERO;
		break;

	case REQ_ADDWHITELIST:
	case REQ_REMWHITELIST:
	case REQ_QUERYWHITELIST:
		if (rp->an < 3 || rp->an > 4) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}

		/* Extract host or network addr */
		if (rp->an == 4)
			cp = rp->av[3];
		else
			cp = NULL;

		/* Extract host or network addr */
		p = extractaddr(rp->av[2], cp, &rp->whitelist);
		if (p != NULL) {
			clientsenderr(cl, rp, "extractaddr: %s", p);
			freereq(rp);
			return;
		}

		p = checkmaskwidth(&rp->whitelist, MAXV4_SIZE, MAXV6_SIZE);
		if (p != NULL) {
			clientsenderr(cl, rp, "%s", p);
			freereq(rp);
			return;
		}

		if (sp->cf->c_whitelistfn == NULL) {
			clientsenderr(cl, rp, "whitelist not configured");
			freereq(rp);
			return;
		}
		break;

	case REQ_LISTACL:
	case REQ_COMPACT:
		if (rp->an != 3) {
			clientsenderr(cl, rp, "wrong number of arguments");
			freereq(rp);
			return;
		}
		rp->aclname = strsave(rp->av[2]);
		break;

	default:
		break;
	}

	/* Validate the port values */
	cp = rp->acl.port1;
	if (cp != NULL && isdigit(*cp)) {
		v = strtol(cp, &ep, 10);
		if (*ep != '\0' || !VALID_PORT(v)) {
			clientsenderr(cl, rp, "bad port valid \"%s\"",
			    cp);
			freereq(rp);
			return;
		}
	}
	cp = rp->acl.port2;
	if (cp != NULL && isdigit(*cp)) {
		v = strtol(cp, &ep, 10);
		if (*ep != '\0' || !VALID_PORT(v)) {
			clientsenderr(cl, rp, "bad port valid \"%s\"",
			    cp);
			freereq(rp);
			return;
		}
	}

	/* Append the new request to end of list */
	if (rp != NULL)
		appendreq(&cl->req, rp);
}

/* Read a command from the client */
void
clientinput(struct state *sp, struct client *cl)
{
	int i;
	char *cp;
	struct req *rp;

	rp = cl->treq;
	if (rp == NULL) {
		/* New request */
		rp = new(1, sizeof(*rp), "clientinput: req");
		cl->treq = rp;
		rp->state = RSTATE_PENDING;

		/* Arrival timestamp */
		getts(&rp->ats);

		cp = iogetstr(&cl->rbuf);
		rp->an = makeargv(cp, &rp->av);

		/* Do we have a comment? */
		if (rp->an > 2 && strcmp(rp->av[rp->an - 1], "-") == 0) {
			--rp->an;
			rp->state = RSTATE_READCOMMENT;

			/* That's all we can do until we collect the comment */
			return;
		}
		/* Fall through and do syntax checks */
	} else {
		/* Request in progress */
		if (rp->state != RSTATE_READCOMMENT) {
			lg(LOG_ERR, "clientinput: unexpected input");
			exit(EX_SOFTWARE);
		}

		/* Collect lines until he's done */
		if (!clientcomment(rp, &cl->rbuf))
			return;

		/* Done reading the comment */
		if (rp->comment != NULL) {
			trimws(rp->comment, strlen(rp->comment));
			lg(LOG_DEBUG, "client #%d comment: \"%s\"",
			    cl->n, pretty(rp->comment, strlen(rp->comment)));
		}
		rp->state = RSTATE_PENDING;
	}

	/* Partial request is now complete */
	cl->treq = NULL;

	if (debug) {
		fprintf(lf, "%s clientinput: client #%d: \"%s",
		    tsstr(), cl->n, rp->an > 0 ? rp->av[0] : "");
		for (i = 1; i < rp->an; ++i)
			fprintf(lf, " %s", rp->av[i]);
		fprintf(lf, "\"\n");
	}

	clientfinishreq(sp, cl, rp);
}

int
clientcomment(struct req *rp, struct iobuf *ip)
{
	char *cp;

	/* Look for end of message (and strip leading dots) */
	while (iohaveline(ip) && (cp = iogetstr(ip)) != NULL) {
		if (*cp == '.') {
			++cp;
			if (strcmp("\r\n", cp) == 0 ||
			    strcmp("\n", cp) == 0 ||
			    *cp == '\0')
				return (1);
		}
		strappend(&rp->comment, cp);
	}
	return (0);
}

void
clientcompact(struct state *sp, struct client *cl, struct req *rp)
{
	struct acllist *ap;
	struct req *rp2;

	ap = aclfindlistbyname(sp->cf, rp->aclname);
	if (ap == NULL) {
		clientsenderr(cl, rp, "acl %s unknown", rp->aclname);
		return;
	}

	if (ap->compactseq > 0 ||
	    ap->compactportseq > 0 ||
	    ap->compactpermithostportseq > 0) {
		rp->flags |= RFLAG_IGNORE;
		clientsendfmt(cl, rp,
		    "compaction already in progress for acl %s", rp->aclname);
		return;
	}

	syslog(LOG_INFO,
	    "clientcompact: started compacting ACL %s (lastseq %d)",
	    ap->name, ap->lastseq);

	/* Compaction only makes sense if we have a range */
	if (sp->cf->c_lowseq != 0 || sp->cf->c_highseq != 0) {
		if (sp->cf->c_incrseq)
			ap->compactseq = sp->cf->c_lowseq;
		else
			ap->compactseq = sp->cf->c_highseq;
	}
	if (sp->cf->c_lowportseq != 0 || sp->cf->c_highportseq != 0)
		ap->compactportseq = sp->cf->c_lowportseq;
	if (sp->cf->c_lowpermithostportseq != 0 ||
	    sp->cf->c_highpermithostportseq != 0)
		ap->compactpermithostportseq = sp->cf->c_lowpermithostportseq;
	ap->compactclient = cl;

	/*
	 * We want to unlink this request but we can't do it here.
	 * So allocate a new one, copy stuff over and mark this one done.
	 */
	rp2 = new(1, sizeof(*rp), "clientcompact: req");
	*rp2 = *rp;
	rp->aclname = NULL;
	rp->comment = NULL;
	rp->payload.buf = NULL;
	rp->payload.len = 0;
	rp->payload.size = 0;
	rp->av = NULL;
	rp->state = RSTATE_DONE;

	rp2->state = RSTATE_PENDING;
	rp2->next = NULL;
	ap->compactreq = rp2;
	timerset(&sp->t_compact, 0);
}

char helpstr[] =
	"drop cookie addr [-]\r\n"
#ifdef notdef
	"drop cookie net/width [-]\r\n"
	"drop cookie net netmask [-]\r\n"
#endif
	"blockhosthost cookie addr addr [-]\r\n"
	"dropudpport cookie port acl [-]\r\n"
	"droptcpport cookie port acl [-]\r\n"
	"droptcpdsthostport cookie addr port [-]\r\n"
	"permitudpdsthostport cookie addr port [-]\r\n"
	"permittcpdsthostport cookie addr port [-]\r\n"
	"restore cookie addr [-]\r\n"
#ifdef notdef
	"restore cookie net/width [-]\r\n"
	"restore cookie net netmask [-]\r\n"
#endif
	"restorehosthost cookie addr addr [-]\r\n"
	"restoretcpdsthostport cookie addr port [-]\r\n"
	"restoretcpport cookie port acl [-]\r\n"
	"restoreudpport cookie port acl [-]\r\n"
	"unpermitudpdsthostport cookie addr port [-]\r\n"
	"unpermittcpdsthostport cookie addr port [-]\r\n"
	"nullzero cookie addr [-]\r\n"
	"nonullzero cookie addr [-]\r\n"
	"addwhitelist cookie addr [-]\r\n"
	"remwhitelist cookie addr [-]\r\n"
	"state cookie\r\n"
	"listacl cookie acl\r\n"
	"listroute cookie\r\n"
	"whitelist cookie\r\n"
	"query cookie addr\r\n"
	"querynullzero cookie addr\r\n"
	"querywhitelist cookie addr\r\n"
	"compact cookie acl\r\n"
	"reload cookie\r\n"
	"exit cookie\r\n"
	"help cookie\r\n";

void
clienthelp(struct client *cl, struct req *rp)
{

	rp->payload.buf = helpstr;
	rp->payload.len = sizeof(helpstr) + 1;
	rp->payload.size = 0;
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

void
clientlistacl(struct state *sp, struct client *cl, struct req *rp)
{
	int i;
	struct acl *al;
	struct acllist *ap;

	/* We only know about configured ACLs */
	ap = aclfindlistbyname(sp->cf, rp->aclname);
	if (ap == NULL) {
		clientsenderr(cl, rp, "acl %s unknown", rp->aclname);
		return;
	}

	for (i = 0, al = ap->acl; i < ap->acllen; ++i, ++al)
		ioappendfmt(&rp->payload, "%s", aclformat(al));
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

void
clientlistroute(struct state *sp, struct client *cl, struct req *rp)
{
	int i;
	struct route *rtp;

	for (i = 0, rtp = sp->routes; i < sp->routeslen; ++i, ++rtp)
		ioappendfmt(&rp->payload, "%s", routeformat(rtp));
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

void
clientprocess(struct state *sp, struct client *cl, struct req *rp)
{
	int i;

	/* First handle requests that never block on talking to the router */
	switch (rp->type) {

	case REQ_UNKNOWN:
		clientsenderr(cl, rp, "unknown command \"%s\"", rp->cmd);
		return;

	case REQ_STATE:
		clientstate(sp, cl, rp);
		return;

	case REQ_EXIT:
		clientsend(cl, rp);
		lg(LOG_DEBUG, "client #%d exit", cl->n);
		/* Flag that this client is finished */
		++cl->close;
		return;

	case REQ_TEST:
		/* Generate lots of output for testing purposes */
		for (i = 1; i <= rp->cookie; ++i)
			ioappendfmt(&rp->payload, "%032d", i);
		rp->flags |= RFLAG_CONTINUE;
		clientsend(cl, rp);
		return;

	case REQ_HELP:
		clienthelp(cl, rp);
		return;

	default:
		/* Fall through */
		break;
	}

	/* If it's not safe to talk to the router, we're done */
	if (!OKTOBLOCK(sp)) {
		/* Be careful not to lock out handleclientreqs() */
		if (cl == sp->reqclient)
			sp->reqclient = NULL;
		return;
	}

	/* Handle potentially blocking requests */
	switch (rp->type) {

	case REQ_RELOAD:
		++reload;
		clientsend(cl, rp);
		break;

	case REQ_DROP:
	case REQ_RESTORE:
	case REQ_BLOCKHOSTHOST:
	case REQ_RESTOREHOSTHOST:
	case REQ_DROPUDPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
	case REQ_DROPTCPDSTHOSTPORT:
	case REQ_RESTORETCPDSTHOSTPORT:
		childdroprestore(sp, cl, rp);
		break;

	case REQ_NULLZERO:
	case REQ_NONULLZERO:
		childnullzero(sp, cl, rp);
		break;

	case REQ_ADDWHITELIST:
	case REQ_REMWHITELIST:
		childwhitelist(sp, cl, rp);
		break;

	case REQ_LISTACL:
		clientlistacl(sp, cl, rp);
		break;

	case REQ_LISTROUTE:
		clientlistroute(sp, cl, rp);
		break;

	case REQ_WHITELIST:
		clientwhitelist(sp, cl, rp);
		break;

	case REQ_QUERY:
		clientquery(sp, cl, rp);
		break;

	case REQ_QUERYNULLZERO:
		clientquerynullzero(sp, cl, rp);
		break;

	case REQ_QUERYWHITELIST:
		clientquerywhitelist(sp, cl, rp);
		break;

	case REQ_COMPACT:
		clientcompact(sp, cl, rp);
		break;

	default:
		lg(LOG_ERR, "clientprocess: unhandled request type %d",
		    rp->type);
		exit(EX_SOFTWARE);
	}
}

void
clientquery(struct state *sp, struct client *cl, struct req *rp)
{
	struct acllist *ap;
	struct acl *al;

	switch (rp->acl.type) {

	case ATYPE_BLOCKHOST:
	case ATYPE_BLOCKNET:
		break;

	default:
		clientsenderr(cl, rp,
		    "don't know how to handle ACL request type %d",
		    rp->acl.type);
		return;
	}

	ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
	if (ap == NULL) {
		clientsenderr(cl, rp,
		    "can't find acllist for %s (is the default ACL missing?)",
		    addr2str(&rp->acl.addr1));
		return;
	}

	al = aclfindacl(ap, &rp->acl);
	if (al == NULL) {
		clientsenderr(cl, rp, "%s not found in ACL %s",
		    addr2str(&rp->acl.addr1), ap->name);
		return;
	}

	clientsendfmt(cl, rp, "acl %s\r\n%s", ap->name, aclformat(al));
}

void
clientquerynullzero(struct state *sp, struct client *cl, struct req *rp)
{
	struct route *rtp;

	if (rp->nullzero.type != ROUTE_NULLZERO) {
		clientsenderr(cl, rp,
		    "don't know how to handle nullzero request type %d",
		    rp->nullzero.type);
		return;
	}

	rtp = routefind(sp, &rp->nullzero);
	if (rtp == NULL) {
		clientsenderr(cl, rp, "%s not found",
		    addr2str(&rp->nullzero.dst));
		return;
	}

	clientsendfmt(cl, rp, "%s", routeformat(rtp));
}

void
clientquerywhitelist(struct state *sp, struct client *cl, struct req *rp)
{
	struct cf *cf;
	struct addr *a;

	cf = sp->cf;
	a = whitelist(cf, &rp->whitelist);
	if (a == NULL) {
		clientsenderr(cl, rp, "%s not found", addr2str(&rp->whitelist));
		return;
	}

	/* Report address or net/mask */
	clientsendfmt(cl, rp, "%s", addr2str(a));
}

/* XXX shouldn't call this until the comment has been collected */
void
clientsend(struct client *cl, struct req *rp)
{
	u_int32_t cookie;
	const char *cmd;
	struct timeval *tvp;
	struct timeval tv;
	char buf[132];

	if (rp != NULL) {
		tvp = &rp->cts;
		if (tvp->tv_sec == 0 && tvp->tv_sec == 0)
			getts(tvp);
		cmd = rp->cmd;
		cookie = rp->cookie;
	} else {
		tvp = &tv;
		getts(tvp);
		cmd = "acld";
		cookie = 0;
	}

#ifdef HAVE_BROCCOLI
	if (cl->broccoli) {
		brosend(cl, rp);
		return;
	}
#endif

	(void)snprintf(buf, sizeof(buf), "%lu.%06lu %u %s%s%s\r\n",
	    (u_long)tvp->tv_sec,
	    (u_long)tvp->tv_usec,
	    cookie,
	    cmd,
	    (rp->flags & RFLAG_FAILED) != 0 ? "-failed" : "",
	    (rp->flags & RFLAG_CONTINUE) != 0 ? " -" : "");
	clientwrite(cl, buf);

	if ((rp->flags & RFLAG_CONTINUE) != 0) {
		if (rp->payload.len > 0)
			clientwrite(cl, rp->payload.buf);
		clientwrite(cl, ".\r\n");
	}

	/* This request is complete */
	rp->state = RSTATE_DONE;
}

void
clientsenderr(struct client *cl, struct req *rp, const char *fmt, ...)
{
	va_list ap;

	rp->flags |= RFLAG_FAILED;
	iofree(&rp->payload);
	va_start(ap, fmt);
	ioappendvfmt(&rp->payload, fmt, ap);
	va_end(ap);
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

void
clientsendfmt(struct client *cl, struct req *rp, const char *fmt, ...)
{
	va_list ap;

	iofree(&rp->payload);
	va_start(ap, fmt);
	ioappendvfmt(&rp->payload, fmt, ap);
	va_end(ap);
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

/* Helper routine for clientstate() */
void
clientcfstate(struct cf *cf, struct req *rp)
{
	int i, j;
	struct acllimit *alp;
	struct acllist *ap;
	struct attrlist *atp;
	struct intlist *ip;
	struct route *rtp;
	struct addr *a;

	if (cf->c_router != NULL)
		ioappendfmt(&rp->payload, "router \"%s\"", cf->c_router);
	if (cf->c_expect != NULL)
		ioappendfmt(&rp->payload, "expect \"%s\"", cf->c_expect);
	if (cf->c_script != NULL)
		ioappendfmt(&rp->payload, "script \"%s\"", cf->c_script);
#ifdef HAVE_CFORCE
	if (cf->c_cforceaddr != NULL)
		ioappendfmt(&rp->payload, "cforceaddr \"%s\"",
		    cf->c_cforceaddr);
	if (cf->c_portcforce != 0)
		ioappendfmt(&rp->payload, "portcforce %d",
		    (int)cf->c_portcforce);
#endif
	if (cf->c_whitelistfn != NULL)
		ioappendfmt(&rp->payload, "whitelist \"%s\"",
		    cf->c_whitelistfn);
	ioappendfmt(&rp->payload, "bindaddr %s", addr2str(&cf->c_bindaddr));
#ifdef HAVE_BROCCOLI
	if (cf->c_portbro != 0 && !ADDREQ(&cf->c_broaddr, &cf->c_bindaddr))
		ioappendfmt(&rp->payload, "broaddr %s",
		    addr2str(&cf->c_broaddr));
#endif
	ioappendfmt(&rp->payload, "port %d", (int)cf->c_port);
#ifdef HAVE_BROCCOLI
	if (cf->c_portbro != 0)
		ioappendfmt(&rp->payload, "portbro %d", (int)cf->c_portbro);
#endif
	if (cf->c_portro != 0)
		ioappendfmt(&rp->payload, "portro %d", (int)cf->c_portro);
	if (cf->c_portweb != 0)
		ioappendfmt(&rp->payload, "portweb %d", (int)cf->c_portweb);
	if (cf->c_netsfac != 0)
		ioappendfmt(&rp->payload, "netsfac %s",
		    val2str(syslog2str, cf->c_netsfac));
	if (cf->c_login_secs != LOGIN_SECS)
		ioappendfmt(&rp->payload, "login_secs %d", cf->c_login_secs);
	if (cf->c_ayt_secs != AYT_SECS)
		ioappendfmt(&rp->payload, "ayt_secs %d", cf->c_ayt_secs);
	if (cf->c_sync_secs != SYNC_SECS)
		ioappendfmt(&rp->payload, "sync_secs %d", cf->c_sync_secs);
	if (cf->c_select_secs != SELECT_SECS)
		ioappendfmt(&rp->payload, "select_secs %d", cf->c_select_secs);
	ioappendfmt(&rp->payload, "incrseq %d", cf->c_incrseq);

	if (cf->c_lowseq != 0 || cf->c_highseq != 0)
		ioappendfmt(&rp->payload, "seqrange %d %d",
		    cf->c_lowseq, cf->c_highseq);
	if (cf->c_maxseq != 0)
		ioappendfmt(&rp->payload, "maxseq %d", cf->c_maxseq);

	if (cf->c_lowportseq != 0 || cf->c_highportseq != 0)
		ioappendfmt(&rp->payload, "portseqrange %d %d",
		    cf->c_lowportseq, cf->c_highportseq);

	if (cf->c_lowpermithostportseq != 0 || cf->c_highpermithostportseq != 0)
		ioappendfmt(&rp->payload, "permithostportseqrange %d %d",
		    cf->c_lowpermithostportseq, cf->c_highpermithostportseq);

	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		ioappendfmt(&rp->payload, "# lastseq for ACL %s is %d",
		    ap->name, ap->lastseq);
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		ioappendfmt(&rp->payload, "# acllen for ACL %s is %d",
		    ap->name, (int)ap->acllen);
	for (i = 0, alp = cf->c_acllimit; i < cf->c_acllimitlen; ++i, ++alp)
		ioappendfmt(&rp->payload, "# limit %d/%d %s", alp->a_len,
		    alp->a_maxseq, fmtargv(alp->a_nacls, alp->a_acls));
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		ioappendfmt(&rp->payload, "# hiwater for ACL %s is %d",
		    ap->name, stats_gethiwater(&ap->stats));
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		ioappendfmt(&rp->payload, "# rate for ACL %s is %d acl/min",
		    ap->name, stats_getrate(&ap->stats));
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap) {
		if (ap->compactseq > 0)
			ioappendfmt(&rp->payload,
			    "# compactseq for ACL %s is %d",
			    ap->name, ap->compactseq);
		if (ap->compactportseq > 0)
			ioappendfmt(&rp->payload,
			    "# compactportseq for ACL %s is %d",
			    ap->name, ap->compactportseq);
		if (ap->compactpermithostportseq > 0)
			ioappendfmt(&rp->payload,
			    "# compactpermithostportseq for ACL %s is %d",
			    ap->name, ap->compactpermithostportseq);
	}
	if (cf->c_whitelistlen != 0)
		ioappendfmt(&rp->payload, "# whitelistlen %d",
		    (int)cf->c_whitelistlen);

	for (i = 0, alp = cf->c_acllimit; i < cf->c_acllimitlen; ++i, ++alp)
		ioappendfmt(&rp->payload, "limit %d %s",
		    alp->a_maxseq, fmtargv(alp->a_nacls, alp->a_acls));
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		for (j = 0, a = ap->addr; j < ap->addrlen; ++j, ++a)
			ioappendfmt(&rp->payload, "acl %s %s",
			    ap->name, addr2str(a));
	for (i = 0, atp = cf->c_attrlist; i < cf->c_attrlistlen; ++i, ++atp)
		ioappendfmt(&rp->payload, "%s", attrfmt(atp));
	for (i = 0, ip = cf->c_intlist; i < cf->c_intlistlen; ++i, ++ip)
		ioappendfmt(&rp->payload, "interface %s %s",
		    ip->i_name, ip->i_acl);
	for (i = 0, rtp = cf->c_nullzeronets; i < cf->c_nullzeronetslen;
	    ++i, ++rtp)
		ioappendfmt(&rp->payload, "nullzeronet %s",
		    addr2str(&rtp->dst));
	if (cf->c_nullzeromax != 0)
		ioappendfmt(&rp->payload, "nullzeromax %d", cf->c_nullzeromax);
}

void
clientstate(struct state *sp, struct client *cl, struct req *rp)
{
	const char *p;
	struct cf *cf;
	char buf[256];

	cf = sp->cf;
	p = val2str(astate2str, sp->state);
	if (p == NULL) {
		(void)snprintf(buf, sizeof(buf), "#%d", sp->state);
		p = buf;
	}
	ioappendfmt(&rp->payload, "# state %s", p);
	ioappendfmt(&rp->payload, "# version %s", version);
	if (gethostname(buf, sizeof(buf)) < 0)
		strcpy(buf, "?");
	ioappendfmt(&rp->payload, "# host %s", buf);
	ioappendfmt(&rp->payload, "# config \"%s\"", configfn);
	if (verbose)
		ioappendfmt(&rp->payload, "# verbose %d", verbose);
	if (debug)
		ioappendfmt(&rp->payload, "# debug %d", debug);
	if (cf != NULL)
		clientcfstate(cf, rp);
	if (sp->nullzerolen != 0)
		ioappendfmt(&rp->payload, "# nullzerolen is %d",
		    sp->nullzerolen);
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

void
clientwhitelist(struct state *sp, struct client *cl, struct req *rp)
{
	int i;
	struct addr *a;
	struct cf *cf;

	cf = sp->cf;
	for (i = 0, a = cf->c_whitelist; i < cf->c_whitelistlen; ++i, ++a)
		ioappendfmt(&rp->payload, "%s", addr2str(a));
	rp->flags |= RFLAG_CONTINUE;
	clientsend(cl, rp);
}

/* Queue up output */
void
clientwrite(struct client *cl, const char *buf)
{
	size_t n;
	ssize_t size;
	struct iobuf *ip;

	ip = &cl->wbuf;
	n = strlen(buf);
	size = ip->len + n + 1;
	DYNARRAY(ip->buf, size, &ip->size,
	    sizeof(*ip->buf), 8192, 8192, "clientwrite iobuf");
	memmove(ip->buf + ip->len, buf, n);
	ip->len += n;
}

void
freeclient(struct state *sp, struct client *cl)
{
	int i;
	struct client *cl2;
	struct req *rp, *rp2;
	struct acllist *ap;

	if (cl->c >= 0) {
		close(cl->c);
		cl->c = -1;
	}
#ifdef HAVE_BROCCOLI
	if (cl->broccoli && cl->bc != NULL) {
		bro_conn_delete(cl->bc);
		cl->bc = NULL;
	}
#endif
	iofree(&cl->rbuf);
	iofree(&cl->wbuf);

	rp = cl->req;
	while (rp != NULL) {
		rp2 = rp->next;
		rp->next = NULL;
		freereq(rp);
		rp = rp2;
	}
	if (cl->treq != NULL)
		freereq(cl->treq);

	if (sp->clients == cl) {
		sp->clients = sp->clients->next;
	} else {
		for (cl2 = sp->clients; cl2 != NULL; cl2 = cl2->next)
			if (cl2->next == cl)
				break;
		if (cl2 == NULL) {
			lg(LOG_ERR, "freeclient: client #%d: can't happen",
			    cl->n);
			exit(EX_SOFTWARE);
		}
		cl2->next = cl->next;
	}
	cl->next = NULL;

	if (cl == sp->reqclient)
		sp->reqclient = NULL;
	free(cl);

	/* Clean up compaction state */
	for (i = 0, ap = sp->cf->c_acllist; i < sp->cf->c_acllistlen; ++i, ++ap)
		if (ap->compactclient == cl) {
			ap->compactclient = NULL;
			if (ap->compactreq != NULL) {
				freereq(ap->compactreq);
				ap->compactreq = NULL;
			}
			break;
		}
}

struct client *
newclient(struct state *sp, int c)
{
	struct client *cl, *cl2;

	cl = (struct client *)new(1, sizeof(*cl), "newclient cl");
	cl->c = c;
	cl->n = sp->nclients++;

	/* Append to end of client list */
	if (sp->clients == NULL) {
		sp->clients = cl;
	} else {
		for (cl2 = sp->clients; cl2 != NULL; cl2 = cl2->next)
			if (cl2->next == NULL)
				break;
		cl2->next = cl;
	}
	return (cl);
}
